@@ -3,6 +3,7 @@ import json
 from datetime import datetime
 from itertools import chain
 
+from py42.response import Py42Response
 from py42.sdk.queries.query_filter import FilterGroup
 from py42.sdk.queries.fileevents.file_event_query import FileEventQuery
 from py42.sdk.queries.fileevents.filters import InsertionTimestamp
@@ -12,14 +13,12 @@ from py42.sdk.queries.alerts.filters.alert_filter import DateObserved
 from c42eventextractor.compat import str
 from c42eventextractor.common import convert_datetime_to_timestamp, IncompatibleFilterError
 
-_MAX_PAGE_SIZE = 10000
 _TIMESTAMP_PRECISION = 0.001
 FILTER_ERROR = u"{} filter can't be used with {} when a cursor checkpoint exists."
 
 
 class BaseExtractor(object):
-
-    _previous_event_count = _MAX_PAGE_SIZE
+    _max_page_size = 10000
 
     def __init__(self, key, search_function, handlers, timestamp_filter, query_class):
         self._key = key
@@ -28,17 +27,13 @@ class BaseExtractor(object):
         self._timestamp_filter = timestamp_filter
         self._sort_key = timestamp_filter._term
         self._query_class = query_class
-
-    @property
-    def previous_event_count(self):
-        return self._previous_event_count
+        self._previous_event_count = self._max_page_size
 
     def extract(self, *args):
-        # type: (iter, iter) -> None
         """Queries for recent events and keeps track of last event timestamp for future filtering.
 
         Passes the raw response from the py42 call to `handlers.handle_response`.
-        The default implementation of `handlers.handle_response` print
+        The default implementation of `handlers.handle_response` prints to the
         console. Provide your own implementation for `handlers.handle_response` to do something
         else. Makes subsequent calls to py42 and `handlers.handle_response` if the total event
         count is greater than 10,000.
@@ -61,60 +56,15 @@ class BaseExtractor(object):
     def extract_advanced(self, query):
         try:
             response = self._search(query)
-            if response.text:
-                self._record_checkpoints(response)
-                self._handlers.handle_response(response)
-            return response
+            return self._handle_response(response)
         except Exception as ex:
             self._handlers.handle_error(ex)
 
-    def _extract_all(self, filter_groups):
-        if self.previous_event_count < _MAX_PAGE_SIZE:
-            return
-
-        original_filters = list(filter_groups)
-        cursor_position = self._handlers.get_cursor_position()
-        if cursor_position:
-            self._add_timestamp_filter(filter_groups, cursor_position)
-        query = self._create_query_from_filters(filter_groups)
-        if self.extract_advanced(query):
-            self._extract_all(original_filters)
-
-    def _record_count(self, response):
-        count = response._data_root.get(u"totalCount", 0)
-        self._previous_event_count = count
-
-    def _record_checkpoints(self, response):
-        self._record_count(response)
-        timestamp = self._extract_timestamp_from_response(response)
-        if timestamp is not None:
-            self._handlers.record_cursor_position(timestamp)
-
-    def _get_results_from_response(self, response):
-        if self._key in response:
-            return response[self._key]
-
-    def _extract_timestamp_from_response(self, response):
-        results = self._get_results_from_response(response)
-        if results and self._timestamp_filter._term in results[-1]:
-            return self._get_timestamp_from_item(results[-1])
-
-    def _create_query_from_filters(self, filter_groups):
-        query = self._query_class(*filter_groups)
-        query.sort_direction = u"asc"
-        query.sort_key = self._sort_key
-        query.page_size = _MAX_PAGE_SIZE
-        return query
-
-    def _add_timestamp_filter(self, filter_groups, cursor_position):
-        timestamp_filter = self._timestamp_filter.on_or_after(
-            cursor_position + _TIMESTAMP_PRECISION
-        )
-        filter_groups.append(timestamp_filter)
-
     def _verify_filter_groups(self, filter_groups, cursor_position):
         if not all(isinstance(group, FilterGroup) for group in filter_groups):
-            raise ValueError("arguments must be py42.sdk.queries.query_filter.FilterGroup objects.")
+            raise ValueError(
+                u"arguments must be py42.sdk.queries.query_filter.FilterGroup objects."
+            )
 
         filters = chain.from_iterable(
             json.loads(str(group)).get(u"filters") for group in filter_groups
@@ -124,11 +74,91 @@ class BaseExtractor(object):
                 FILTER_ERROR.format(self._timestamp_filter.__name__, self.__class__.__name__)
             )
 
+    def _extract_all(self, filter_groups):
+        if self._previous_event_count < self._max_page_size:
+            return
+
+        original_filters = list(filter_groups)
+        cursor_position = self._handlers.get_cursor_position()
+
+        if cursor_position:
+            self._add_timestamp_filter(filter_groups, cursor_position)
+
+        query = self._create_query_from_filters(filter_groups)
+
+        # do not recurse if handler returns no cursor
+        if self.extract_advanced(query) and self._handlers.get_cursor_position():
+            self._extract_all(original_filters)
+
+    def _add_timestamp_filter(self, filter_groups, cursor_position):
+        timestamp_filter = self._timestamp_filter.on_or_after(cursor_position)
+        filter_groups.append(timestamp_filter)
+
+    def _create_query_from_filters(self, filter_groups):
+        query = self._query_class(*filter_groups)
+        query.sort_direction = u"asc"
+        query.sort_key = self._sort_key
+        query.page_size = self._max_page_size
+        return query
+
+    def _get_results_from_response(self, response):
+        return response[self._key]
+
+    def _handle_response(self, response):
+        results = self._get_results_from_response(response)
+        if results and self._timestamp_filter._term in results[-1]:
+            timestamp, index = self._extract_timestamp_from_results(results)
+            self._record_count(response)
+            if index and self._previous_event_count >= self._max_page_size:
+                # truncate responses with the highest timestamp to avoid duplicates
+                response._data_root[self._key] = response[self._key][:index]
+            else:
+                # either this is the last page, OR
+                # every item on this page had the same timestamp.
+                # increase the timestamp by .001 and continue
+                timestamp += _TIMESTAMP_PRECISION
+
+            self._handlers.handle_response(response)
+            self._record_checkpoint(timestamp)
+            return response
+
+    def _extract_timestamp_from_results(self, results):
+        """Return the highest timestamp found by the query and its index.
+         This info is used to continue finding results without recording any duplicates.
+        """
+        index = self._get_index_where_truncation_begins(results)
+        highest_timestamp = self._get_timestamp_from_item(results[index])
+
+        return highest_timestamp, index
+
+    def _get_index_where_truncation_begins(self, results):
+        current_index = 0
+        index = 0
+        cutoff = self._get_timestamp_from_item(results[-1])
+        results = results[:-1]
+        for result in reversed(results):
+            current_index -= 1
+            next_most_recent = self._get_timestamp_from_item(result)
+            if next_most_recent < cutoff:
+                index = current_index
+                break
+        return index
+
+    def _record_checkpoint(self, timestamp):
+        if timestamp is not None:
+            self._handlers.record_cursor_position(timestamp)
+
+    def _record_count(self, response):
+        count = response._data_root.get(u"totalCount", 0)
+        self._previous_event_count = count
+
     def _get_timestamp_from_item(self, item):
         raise NotImplementedError
 
 
 class AlertExtractor(BaseExtractor):
+    _max_page_size = 500
+
     def __init__(self, sdk, handlers):
         super(AlertExtractor, self).__init__(
             key=u"alerts",
